#
# ## Mutual information as an image matching metric
#
# In which we look at the mutual information measure for comparing images.
#
# [LICENSE](https://matthew-brett.github.io/teaching/_downloads/15e073449227fc6a31bb4db45d795839/LICENSE) file for copyright.
# Script by Matthew Brett. Adapted for BrainVoyager 05-02-2020

# - compatibility with Python 2
from __future__ import print_function  # print('me') instead of print 'me'
from __future__ import division  # 1/2 == 0.5, not 0

import PythonQt.Qt as qt
import os
import scipy as sp
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg

# - import common modules
import numpy as np  # the Python array package
import matplotlib.pyplot as plt  # the Python plotting package
from scipy import ndimage, misc 

import pylab
#import bv

fig1 = Figure(figsize=(5, 3)) # from Matplotlib
ax1 = fig1.add_subplot(111) 

def create_fig_dlg(fig, title):    
    canvas = FigureCanvasAgg(fig)
    canvas.print_figure(os.path.join(os.getcwd(), '.matplotlib-for-qt.png')) # current solution: save matplotlib-fig to disk and reload as QImage below
    img = qt.QImage()
    ok = img.load(os.path.join(os.getcwd(), '.matplotlib-for-qt.png'))
    if ok == False:
        return
    dlg = qt.QWidget()
    dlg.setAttribute(55, True) # 55 -> WA_DeleteOnClose
    #dlg.setWindowTitle("matplotlib in Qt GUI")
    dlg.setWindowTitle(title)
    box = qt.QVBoxLayout(dlg)
    label = qt.QLabel(dlg)
    label.setContentsMargins(0, 0, 0, 0)
    box.setContentsMargins(0, 0, 0, 0)
    box.addWidget(label)
    pxm = qt.QPixmap()
    pxm.convertFromImage(img)
    label.setPixmap(pxm)
    label.setScaledContents(True)
    return dlg
#~~~~
#canvas = FigureCanvasAgg(fig)
#canvas.print_figure("test.png", dpi=200) 
#~~~~


# - set gray colormap and nearest neighbor interpolation by default
plt.rcParams['image.cmap'] = 'gray'
plt.rcParams['image.interpolation'] = 'nearest'

BrainVoyager.PrintToLog('This script will calculate mutual information of two slices of VMR files.\nScript by Matthew Brett (https://matthew-brett.github.io/.')

# get the anatomical BV data
vmrfilename = BrainVoyager.BrowseFile("Please select an VMR File", "*.vmr")
docvmr = BrainVoyager.OpenDocument(vmrfilename)
volumeData = docvmr.GetVolumeDataAsDoubleList()
t1_data1D = np.array(volumeData)                       # convert list (or tuple) to numpy array
t1_data = t1_data1D.reshape(docvmr.DimX, docvmr.DimY, docvmr.DimZ) # reshape 1D to 3D array
BrainVoyager.PrintToLog('First file:\n' + vmrfilename)

# get the functional data (VTC-in-VMR via FMR -> VTC -> Show VTC Volume (3D Volume Tools) -> Save secondary VMR)
vmrfilename = BrainVoyager.BrowseFile("Please select the FMR in VMR File", "*.vmr")
docvmr2 = BrainVoyager.OpenDocument(vmrfilename)
BrainVoyager.PrintToLog('Second file:\n' + vmrfilename)
volumeData2 = docvmr2.GetVolumeDataAsDoubleList()
t2_data1D = np.array(volumeData2)                       # convert list (or tuple) to numpy array
t2_data = t2_data1D.reshape(docvmr.DimX, docvmr.DimY, docvmr.DimZ) # reshape 1D to 3D array

# Show the images by stacking them left-right with hstack
t1_slice = t1_data[:, :, 128]
t1_slice90 = ndimage.rotate(t1_slice, -90, reshape=False)
t2_slice = t2_data[:, :, 128]
t2_slice90 = ndimage.rotate(t2_slice, -90, reshape=False)
ax1.imshow(np.hstack((t1_slice90, t2_slice90))) 
dlg1 = create_fig_dlg(fig1, 'T1 and T2') # specific for BV21
dlg1.show() 

# The one-dimensional histograms of the example slices:
fig2 = Figure(figsize=(9, 5)) # matplotlib Figure
ax2 = fig2.subplots(1,2) 
ax2[0].hist(t1_slice90.ravel(), bins=20) 
ax2[0].set_title('T1 slice histogram')
ax2[1].hist(t2_slice90.ravel(), bins=20)
ax2[1].set_title('T2 slice histogram')
dlg2 = create_fig_dlg(fig2, 'T1 and T2 histograms') # specific for BV21 
dlg2.show() 

# Plotting the signal in the T1 slice against the signal in the T2 slice:
fig3 = Figure(figsize=(5, 5)) # from Matplotlib
ax3 = fig3.add_subplot(111) 
ax3.plot(t1_slice90.ravel(), t2_slice90.ravel(), '.')
ax3.set_xlabel('T1 signal') 
ax3.set_ylabel('T2 signal')
np.corrcoef(t1_slice90.ravel(), t2_slice90.ravel())[0, 1]
ax3.set_title('T1 vs T2 signal')
dlg3 = create_fig_dlg(fig3, 'Signal of T1 against T2') # specific for BV21 
dlg3.show() 
print('Correlation: ' + str(np.corrcoef(t1_slice90.ravel(), t2_slice90.ravel())[0, 1]))
BrainVoyager.PrintToLog('Correlation: ' + str(np.corrcoef(t1_slice90.ravel(), t2_slice90.ravel())[0, 1]) ) 


# Notice that we can predict the T2 signal given the T1 signal, but it is not a
# linear relationship. For example, for T1 signal between 20 and 30, most
# corresponding T2 signal is low, but there is some T2 signal that is high. This
# pairing of high T2 signal with low T1 signal is from the CSF, which is dark
# (low signal) in the T1, and bright in the T2.

# Array that is True if T1 signal >= 20, <= 30, False otherwise
#t1_20_30 = (t1_slice >= 20) & (t1_slice <= 30) # using VTC-> VMR so intensities are modified
t1_20_30_90 = (t1_slice90 >= 150) & (t1_slice90 <= 170)
fig4 = Figure(figsize=(8, 3)) # matplotlib Figure
ax4 = fig4.subplots(1,3) 
ax4[0].imshow(t1_slice90) 
ax4[0].set_title('T1 slice')
ax4[1].imshow(t1_20_30_90)
ax4[1].set_title('150<=T1<=170')
ax4[2].imshow(t2_slice90)
ax4[2].set_title('T2 slice')
dlg4 = create_fig_dlg(fig4, 'T1 and T2 intensity differences') # specific for BV21 
dlg4.show() 


# We can capture this more complicated relationship by doing a 2D histogram.
# This is a histogram that divides the scatterplot into squares, and counts the
# number of observations inside each square.
#
# Look again at the [scatterplot for the T1 and T2 values](https://matthew-brett.github.io/teaching/mutual_information.html#t1-t2-scatter).
# We get the 1D histogram for T1 values by splitting the x axis into bins, and
# taking the number of observations contained in each column defined by the
# bins.  The T2 histogram comes from splitting the y axis into bins and taking
# the number of observations contained in each row defined by the bins.  The 2D
# histogram comes from dividing both the x and the y axis into bins and taking
# the number of observations in each square defined by the intersection of the
# rows and columns:
#
# Numpy has a function for doing the 2D histogram calculation:

hist_2d, x_edges, y_edges = np.histogram2d(
    t1_slice.ravel(),
    t2_slice.ravel(),
    bins=20)
# Plot as image, arranging axes as for scatterplot
# We transpose to put the T1 bins on the horizontal axis
# and use 'lower' to put 0, 0 at the bottom of the plot
fig5 = Figure(figsize=(5, 5)) # from Matplotlib
ax5 = fig5.add_subplot(111) 
ax5.imshow(hist_2d.T, origin='lower')
#ax5.plot(t1_slice.ravel(), t2_slice.ravel(), '.')
ax5.set_xlabel('T1 signal bin') 
ax5.set_ylabel('T2 signal bin')
dlg5 = create_fig_dlg(fig5, '2D histogram with 20 bins') # specific for BV21  
dlg5.show()

# The histogram is easier to see if we show the log values to reduce the effect
# of the bins with a very large number of values:

# Show log histogram, avoiding divide by 0
hist_2d_log = np.zeros(hist_2d.shape)
non_zeros = hist_2d != 0
hist_2d_log[non_zeros] = np.log(hist_2d[non_zeros])
fig6 = Figure(figsize=(6, 6)) # from Matplotlib
ax6 = fig6.add_subplot(111) 
ax6.imshow(hist_2d_log.T, origin='lower')
ax6.set_xlabel('T1 signal bin') 
ax6.set_ylabel('T2 signal bin')
ax6.set_title('Mutual information')
dlg6 = create_fig_dlg(fig6, '2D log histogram') # specific for BV21 
dlg6.show()  

# Mutual information is a metric from the joint (2D) histogram. The metric is
# high when the signal is highly concentrated in few bins (squares), and low
# when the signal is spread across many bins (squares).
#
# Mutual information is defined as:
#
# $$
# I(X;Y) = \sum_{y \in Y} \sum_{x \in X}
#                  p(x,y) \log{ \left(\frac{p(x,y)}{p(x)\,p(y)}
#                               \right) }
# $$
#
# See [http://en.wikipedia.org/wiki/Mutual_information](http://en.wikipedia.org/wiki/Mutual_information)

def mutual_information(hgram):
    """ Mutual information for joint histogram
    """
    # Convert bins counts to probability values
    pxy = hgram / float(np.sum(hgram))
    px = np.sum(pxy, axis=1) # marginal for x over y
    py = np.sum(pxy, axis=0) # marginal for y over x
    px_py = px[:, None] * py[None, :] # Broadcast to multiply marginals
    # Now we can do the calculation using the pxy, px_py 2D arrays
    nzs = pxy > 0 # Only non-zero pxy values contribute to the sum
    return np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))

print('Mutual information: ' + str(mutual_information(hist_2d)))
BrainVoyager.PrintToLog('Mutual information: ' + str(mutual_information(hist_2d)))

# If we move the T2 image 15 pixels down, we make the images less well
# registered.
t2_slice_moved90 = np.zeros(t2_slice90.shape)
t2_slice_moved90[15:, :] = t2_slice90[:-15, :]
fig7 = Figure(figsize=(4, 4)) # from Matplotlib
ax7 = fig7.add_subplot(111) 
ax7.imshow(np.hstack((t1_slice90, t2_slice_moved90)))
dlg7 = create_fig_dlg(fig7, 'T2 slice moved') # specific for BV21  
dlg7.show() 

# Now the scatterplot is a lot more diffuse:
fig8 = Figure(figsize=(5, 5)) # from Matplotlib
ax8 = fig8.add_subplot(111) 
ax8.plot(t1_slice90.ravel(), t2_slice_moved90.ravel(), '.')
ax8.set_xlabel('T1 signal') 
ax8.set_ylabel('T2 (moved) signal')
ax8.set_title('T1 vs T2 signal after moving T2 image')
dlg8 = create_fig_dlg(fig8, 'T1 vs T2 signal after moving T2 image') # specific for BV21 
dlg8.show()  
#print('Correlation (moved T2): ' + str(np.corrcoef(t1_slice90.ravel(), t2_slice_moved90.ravel())[0, 1]))
#BrainVoyager.PrintToLog('Correlation (moved T2): ' + str(np.corrcoef(t1_slice90.ravel(), t2_slice_moved90.ravel())[0, 1]) )

# The joint (2D) histogram shows the same thing:
hist_2d_moved, x_edges, y_edges = np.histogram2d(
    t1_slice90.ravel(),
    t2_slice_moved90.ravel(),
    bins=20)

# Show log histogram, avoiding divide by 0
hist_2d_moved_log = np.zeros(hist_2d_moved.shape)
non_zeros = hist_2d_moved != 0
hist_2d_moved_log[non_zeros] = np.log(hist_2d_moved[non_zeros])
non_zeros = hist_2d != 0
hist_2d_log[non_zeros] = np.log(hist_2d[non_zeros])
fig9 = Figure(figsize=(5, 5)) # from Matplotlib
ax9 = fig9.add_subplot(111) 
ax9.imshow(hist_2d_moved_log.T, origin='lower')
ax9.set_xlabel('T1 signal bin') 
ax9.set_ylabel('T2 signal bin')
mutual_information(hist_2d_moved)
dlg9 = create_fig_dlg(fig9, '2D log histogram (moved T2)') # specific for BV21 
dlg9.show()  

# Because the signal is less concentrated into a small number of bins, the
# mutual information has dropped:
print('Mutual information (moved T2): ' + str(mutual_information(hist_2d_moved)))
BrainVoyager.PrintToLog('Mutual information (moved T2): ' + str(mutual_information(hist_2d_moved)) )

